"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _neoBlockchainNetwork = require("neo-blockchain-network");

var _neoBlockchainCore = require("neo-blockchain-core");

var _neoBlockchainNodeCore = require("neo-blockchain-node-core");

var _bloomFilter = _interopRequireDefault(require("bloom-filter"));

var _bloem = require("bloem");

var _lodash = _interopRequireDefault(require("lodash"));

var _net = _interopRequireDefault(require("net"));

var _Command = require("./Command");

var _payload = require("./payload");

var _Message = _interopRequireWildcard(require("./Message"));

var _errors = require("./errors");

var _package = _interopRequireDefault(require("../package.json"));

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _extends() { _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; }; return _extends.apply(this, arguments); }

const createPeerBloomFilter = ({
  filter,
  k,
  tweak
}) => new _bloomFilter.default({
  vData: Buffer.from(filter),
  nHashFuncs: k,
  nTweak: tweak
});

const createScalingBloomFilter = () => new _bloem.ScalingBloem(0.05, {
  initial_capacity: 100000,
  scaling: 4
});

const MEM_POOL_SIZE = 30000;
const GET_ADDR_PEER_COUNT = 200;
const GET_BLOCKS_COUNT = 500; // Assume that we get 500 back, but if not, at least request every 10 seconds

const GET_BLOCKS_BUFFER = GET_BLOCKS_COUNT / 3;
const GET_BLOCKS_TIME_MS = 10000;
const GET_BLOCKS_THROTTLE_MS = 500;
const GET_BLOCKS_CLOSE_COUNT = 3;
const LOCAL_HOST_ADDRESSES = new Set(['0.0.0.0', 'localhost', '127.0.0.1']);

class Node {
  constructor(options) {
    Object.defineProperty(this, "_blockchain", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_network", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_started", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_stopped", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_externalPort", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_nonce", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_userAgent", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "memPool", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_knownBlockHashes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_tempKnownBlockHashes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_knownTransactionHashes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_tempKnownTransactionHashes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_knownHeaderHashes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_tempKnownHeaderHashes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_fetchedIndex", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_getBlocksRequestsIndex", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_getBlocksRequestTime", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_getBlocksRequestsCount", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_bestPeer", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "_negotiate", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: async peer => {
        this._sendMessage(peer, this._createMessage({
          command: _Command.COMMAND.VERSION,
          payload: new _payload.VersionPayload({
            protocolVersion: 0,
            services: _payload.SERVICES.NODE_NETWORK,
            timestamp: Math.round(Date.now() / 1000),
            port: this._externalPort,
            nonce: this._nonce,
            userAgent: this._userAgent,
            startHeight: this._blockchain.currentBlockIndex,
            relay: true
          })
        }));

        const message = await peer.receiveMessage(30000);
        let versionPayload;

        if (message.value.command === _Command.COMMAND.VERSION) {
          versionPayload = message.value.payload;
        } else {
          throw new _errors.NegotiationError(message);
        }

        if (!this._checkVersion(versionPayload)) {
          throw new _errors.NegotiationError(message);
        }

        let address;
        const {
          host
        } = (0, _neoBlockchainNodeCore.getEndpointConfig)(peer.endpoint);

        if (_net.default.isIPv4(host)) {
          address = new _payload.NetworkAddress({
            host,
            port: versionPayload.port,
            timestamp: versionPayload.timestamp,
            services: versionPayload.services
          });
        }

        this._sendMessage(peer, this._createMessage({
          command: _Command.COMMAND.VERACK
        }));

        const nextMessage = await peer.receiveMessage(30000);

        if (nextMessage.value.command !== _Command.COMMAND.VERACK) {
          throw new _errors.NegotiationError(nextMessage);
        }

        return {
          data: {
            nonce: versionPayload.nonce,
            startHeight: versionPayload.startHeight,
            bloomFilter: null,
            address
          },
          relay: versionPayload.relay
        };
      }
    });
    Object.defineProperty(this, "_onEvent", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: event => {
        if (event.event === 'PEER_CONNECT_SUCCESS') {
          const {
            connectedPeer
          } = event.extra;

          if (this._bestPeer == null || this._bestPeer.data.startHeight < connectedPeer.data.startHeight) {
            this._bestPeer = connectedPeer;

            this._requestBlocks();
          }
        } else if (event.event === 'PEER_CLOSED') {
          if (this._bestPeer === event.extra.peer) {
            this._bestPeer = this._findBestPeer();

            this._requestBlocks();
          }
        }

        this._blockchain.log(_extends({
          event: event.event
        }, event.data || {}));
      }
    });
    Object.defineProperty(this, "_requestBlocks", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: _lodash.default.throttle(_lodash.default.debounce(() => {
        const peer = this._bestPeer;
        const block = this._blockchain.currentBlock;

        if (peer != null && block.index < peer.data.startHeight) {
          if (this._getBlocksRequestsCount > GET_BLOCKS_CLOSE_COUNT) {
            peer.close();
            this._getBlocksRequestsCount = 0;
          } else if (this._shouldRequestBlocks()) {
            if (this._getBlocksRequestsIndex === block.index) {
              this._getBlocksRequestsCount += 1;
            } else {
              this._getBlocksRequestsCount = 1;
              this._getBlocksRequestsIndex = block.index;
            }

            this._getBlocksRequestTime = Date.now();

            this._sendMessage(peer, this._createMessage({
              command: _Command.COMMAND.GET_BLOCKS,
              payload: new _payload.GetBlocksPayload({
                hashStart: [block.hash]
              })
            }));
          }

          this._requestBlocks();
        }
      }, GET_BLOCKS_THROTTLE_MS), GET_BLOCKS_THROTTLE_MS)
    });
    Object.defineProperty(this, "_onRequestEndpoints", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: () => {
        this._relay(this._createMessage({
          command: _Command.COMMAND.GET_ADDR
        }));
      }
    });
    this._blockchain = options.blockchain;
    this._network = new _neoBlockchainNetwork.Network({
      seeds: options.seeds,
      listenTCP: options.listenTCP,
      externalEndpoints: options.externalEndpoints,
      connectPeersDelayMS: options.connectPeersDelayMS,
      maxConnectedPeers: options.maxConnectedPeers,
      socketTimeoutMS: options.socketTimeoutMS,
      negotiate: this._negotiate,
      createMessageTransform: () => new _Message.MessageTransform(this._blockchain.deserializeWireContext),
      onMessageReceived: this._onMessageReceived.bind(this),
      onRequestEndpoints: this._onRequestEndpoints.bind(this),
      onEvent: this._onEvent
    });
    this._started = false;
    this._stopped = false;
    this._externalPort = (options.listenTCP || {}).port || 0;
    this._nonce = Math.floor(Math.random() * _neoBlockchainCore.utils.UINT_MAX_NUMBER);
    this._userAgent = `NEO:neo-blockchain-js:${_package.default.version}`;
    this.memPool = {};
    this._knownBlockHashes = createScalingBloomFilter();
    this._tempKnownBlockHashes = new Set();
    this._knownTransactionHashes = createScalingBloomFilter();
    this._tempKnownTransactionHashes = new Set();
    this._knownHeaderHashes = createScalingBloomFilter();
    this._tempKnownHeaderHashes = new Set();
    this._getBlocksRequestsIndex = null;
    this._getBlocksRequestTime = null;
    this._getBlocksRequestsCount = 1;
  }

  get connectedPeersCount() {
    return this._network.connectedPeers.length;
  }

  start() {
    if (this._started) {
      return;
    }

    this._started = true;
    this._stopped = false;

    this._blockchain.log({
      event: 'NODE_START'
    });

    try {
      this._network.start();
    } catch (error) {
      this._started = false;
      throw error;
    }
  }

  stop() {
    if (this._stopped) {
      return;
    }

    this._stopped = true;

    this._blockchain.log({
      event: 'NODE_STOP'
    });

    try {
      this._network.stop();

      this._started = false;
    } catch (error) {
      this._stopped = false;
      throw error;
    }
  }

  async relayTransaction(transaction) {
    if (transaction.type === _neoBlockchainCore.TRANSACTION_TYPE.MINER || this.memPool[transaction.hashHex] != null || this._tempKnownTransactionHashes.has(transaction.hashHex)) {
      return;
    }

    if (!this._knownTransactionHashes.has(transaction.hash)) {
      const hash = _neoBlockchainCore.common.uInt256ToString(transaction.hash);

      this._blockchain.log({
        event: 'RELAY_TRANSACTION_START',
        level: 'debug',
        hash
      });

      this._tempKnownTransactionHashes.add(transaction.hashHex);

      try {
        const foundTransaction = await this._blockchain.transaction.tryGet({
          hash: transaction.hash
        });

        if (foundTransaction == null) {
          await this._blockchain.verifyTransaction({
            transaction,
            memPool: _neoBlockchainCore.utils.values(this.memPool)
          });
          this.memPool[transaction.hashHex] = transaction;

          this._knownTransactionHashes.add(transaction.hash);

          this._relayTransaction(transaction);

          this._blockchain.log({
            event: 'RELAY_TRANSACTION_SUCCESS',
            level: 'debug',
            hash
          });

          await this._trimMemPool();
        }
      } catch (error) {
        this._blockchain.log({
          event: 'RELAY_TRANSACTION_ERROR',
          hash,
          error
        });

        throw error; // eslint-disable-next-line
      } finally {
        this._tempKnownTransactionHashes.delete(transaction.hashHex);
      }
    }
  }

  _relay(message) {
    this._blockchain.log({
      event: 'RELAY_MESSAGE',
      level: 'debug',
      command: message.value.command
    });

    this._network.relay(message.serializeWire());
  }

  _relayTransaction(transaction) {
    const message = this._createMessage({
      command: _Command.COMMAND.INV,
      payload: new _payload.InvPayload({
        type: _payload.INVENTORY_TYPE.TRANSACTION,
        hashes: [transaction.hash]
      })
    });

    const messagePayload = message.serializeWire();

    for (const peer of this._network.connectedPeers) {
      if (peer.relay && this._testFilter(peer.data.bloomFilter, transaction)) {
        peer.write(messagePayload);
      }
    }
  }

  _sendMessage(peer, message) {
    this._blockchain.log({
      event: 'SEND_MESSAGE',
      level: 'debug',
      command: message.value.command,
      endpoint: peer.endpoint
    });

    peer.write(message.serializeWire());
  }

  _findBestPeer() {
    return _lodash.default.maxBy(this._network.connectedPeers, peer => peer.data.startHeight);
  }

  _shouldRequestBlocks() {
    const block = this._blockchain.currentBlock;
    const getBlocksRequestTime = this._getBlocksRequestTime;
    return this._getBlocksRequestsIndex == null || block.index - this._getBlocksRequestsIndex > GET_BLOCKS_BUFFER || getBlocksRequestTime == null || Date.now() - getBlocksRequestTime > GET_BLOCKS_TIME_MS;
  }

  _checkVersion(version) {
    if (version.nonce === this._nonce) {
      return false;
    }

    if (this._network.connectedPeers.some(peer => version.nonce === peer.data.nonce)) {
      return false;
    }

    return true;
  }

  _onMessageReceived(peer, message) {
    this._blockchain.log({
      event: 'MESSAGE_RECEIVED',
      level: 'debug',
      endpoint: peer.endpoint,
      command: message.value.command
    });

    const onError = error => {
      this._blockchain.log({
        event: 'MESSAGE_RECEIVED_ERROR',
        error,
        endpoint: peer.endpoint,
        command: message.value.command
      });
    };

    try {
      switch (message.value.command) {
        case _Command.COMMAND.ADDR:
          this._onAddrMessageReceived(message.value.payload);

          break;

        case _Command.COMMAND.BLOCK:
          this._onBlockMessageReceived(message.value.payload).catch(onError);

          break;

        case _Command.COMMAND.CONSENSUS:
          this._onConsensusMessageReceived(message.value.payload);

          break;

        case _Command.COMMAND.FILTER_ADD:
          this._onFilterAddMessageReceived(peer, message.value.payload);

          break;

        case _Command.COMMAND.FILTER_CLEAR:
          this._onFilterClearMessageReceived(peer);

          break;

        case _Command.COMMAND.FILTER_LOAD:
          this._onFilterLoadMessageReceived(peer, message.value.payload);

          break;

        case _Command.COMMAND.GET_ADDR:
          this._onGetAddrMessageReceived(peer);

          break;

        case _Command.COMMAND.GET_BLOCKS:
          this._onGetBlocksMessageReceived(peer, message.value.payload).catch(onError);

          break;

        case _Command.COMMAND.GET_DATA:
          this._onGetDataMessageReceived(peer, message.value.payload).catch(onError);

          break;

        case _Command.COMMAND.GET_HEADERS:
          this._onGetHeadersMessageReceived(peer, message.value.payload).catch(onError);

          break;

        case _Command.COMMAND.HEADERS:
          this._onHeadersMessageReceived(peer, message.value.payload).catch(onError);

          break;

        case _Command.COMMAND.INV:
          this._onInvMessageReceived(peer, message.value.payload);

          break;

        case _Command.COMMAND.MEMPOOL:
          this._onMemPoolMessageReceived(peer);

          break;

        case _Command.COMMAND.TX:
          this._onTransactionReceived(message.value.payload).catch(onError);

          break;

        case _Command.COMMAND.VERACK:
          this._onVerackMessageReceived(peer);

          break;

        case _Command.COMMAND.VERSION:
          this._onVersionMessageReceived(peer);

          break;

        case _Command.COMMAND.ALERT:
          break;

        case _Command.COMMAND.MERKLE_BLOCK:
          break;

        case _Command.COMMAND.NOT_FOUND:
          break;

        case _Command.COMMAND.PING:
          break;

        case _Command.COMMAND.PONG:
          break;

        case _Command.COMMAND.REJECT:
          break;

        default:
          // eslint-disable-next-line
          message.value.command;
          break;
      }
    } catch (error) {
      onError(error);
    }
  }

  _onAddrMessageReceived(addr) {
    addr.addresses.filter(address => !LOCAL_HOST_ADDRESSES.has(address.host)).map(address => (0, _neoBlockchainNodeCore.createEndpoint)({
      type: 'tcp',
      host: address.host,
      port: address.port
    })).forEach(endpoint => this._network.addEndpoint(endpoint));
  }

  async _onBlockMessageReceived(block) {
    if (this._blockchain.currentBlockIndex >= block.index || this._tempKnownBlockHashes.has(block.hashHex)) {
      return;
    }

    if (!this._knownBlockHashes.has(block.hash)) {
      this._tempKnownBlockHashes.add(block.hashHex);

      try {
        const foundBlock = await this._blockchain.block.tryGet({
          hashOrIndex: block.hash
        });

        if (foundBlock == null) {
          await this._blockchain.persistBlock({
            block
          });
          const peer = this._bestPeer;

          if (peer != null && block.index > peer.data.startHeight) {
            this._relay(this._createMessage({
              command: _Command.COMMAND.INV,
              payload: new _payload.InvPayload({
                type: _payload.INVENTORY_TYPE.BLOCK,
                hashes: [block.hash]
              })
            }));
          }
        }

        this._knownBlockHashes.add(block.hash);

        this._knownHeaderHashes.add(block.hash);

        for (const transaction of block.transactions) {
          delete this.memPool[transaction.hashHex];

          this._knownTransactionHashes.add(transaction.hash);
        }
      } finally {
        this._tempKnownBlockHashes.delete(block.hashHex);
      }
    }
  } // TODO: Implement
  // eslint-disable-next-line


  _onConsensusMessageReceived(consensus) {} // eslint-disable-next-line


  _onFilterAddMessageReceived(peer, filterAdd) {
    if (peer.data.bloomFilter != null) {
      peer.data.bloomFilter.insert(filterAdd.data);
    }
  } // eslint-disable-next-line


  _onFilterClearMessageReceived(peer) {
    // eslint-disable-next-line
    peer.data.bloomFilter = null;
  } // eslint-disable-next-line


  _onFilterLoadMessageReceived(peer, filterLoad) {
    // eslint-disable-next-line
    peer.data.bloomFilter = createPeerBloomFilter(filterLoad);
  }

  _onGetAddrMessageReceived(peer) {
    const addresses = _lodash.default.take(_lodash.default.shuffle(this._network.connectedPeers.map(connectedPeer => connectedPeer.data.address).filter(Boolean)), GET_ADDR_PEER_COUNT);

    if (addresses.length > 0) {
      this._sendMessage(peer, this._createMessage({
        command: _Command.COMMAND.ADDR,
        payload: new _payload.AddrPayload({
          addresses
        })
      }));
    }
  }

  async _onGetBlocksMessageReceived(peer, getBlocks) {
    const headers = await this._getHeaders(getBlocks, this._blockchain.currentBlockIndex);

    this._sendMessage(peer, this._createMessage({
      command: _Command.COMMAND.INV,
      payload: new _payload.InvPayload({
        type: _payload.INVENTORY_TYPE.BLOCK,
        hashes: headers.map(header => header.hash)
      })
    }));
  }

  async _onGetDataMessageReceived(peer, getData) {
    switch (getData.type) {
      case 0x01:
        // Transaction
        await Promise.all(getData.hashes.map(async hash => {
          let transaction = this.memPool[_neoBlockchainCore.common.uInt256ToHex(hash)];

          if (transaction == null) {
            transaction = await this._blockchain.transaction.tryGet({
              hash
            });
          }

          if (transaction != null) {
            this._sendMessage(peer, this._createMessage({
              command: _Command.COMMAND.TX,
              payload: transaction
            }));
          }
        }));
        break;

      case 0x02:
        // Block
        await Promise.all(getData.hashes.map(async hash => {
          const block = await this._blockchain.block.tryGet({
            hashOrIndex: hash
          });

          if (block != null) {
            if (peer.data.bloomFilter == null) {
              this._sendMessage(peer, this._createMessage({
                command: _Command.COMMAND.BLOCK,
                payload: block
              }));
            } else {
              this._sendMessage(peer, this._createMessage({
                command: _Command.COMMAND.MERKLE_BLOCK,
                payload: this._createMerkleBlockPayload({
                  block,
                  flags: block.transactions.map(transaction => this._testFilter(peer.data.bloomFilter, transaction))
                })
              }));
            }
          }
        }));
        break;

      case 0xe0:
        // Consensus
        // TODO: Implement
        break;

      default:
        // eslint-disable-next-line
        getData.type;
        break;
    }
  }

  async _onGetHeadersMessageReceived(peer, getBlocks) {
    const headers = await this._getHeaders(getBlocks, this._blockchain.currentHeader.index);

    this._sendMessage(peer, this._createMessage({
      command: _Command.COMMAND.HEADERS,
      payload: new _payload.HeadersPayload({
        headers
      })
    }));
  }

  async _onHeadersMessageReceived(peer, headersPayload) {
    const headers = headersPayload.headers.filter(header => !this._knownHeaderHashes.has(header.hash) && !this._tempKnownHeaderHashes.has(header.hashHex));

    if (headers.length > 0) {
      for (const header of headers) {
        this._tempKnownHeaderHashes.add(header.hashHex);
      }

      try {
        await this._blockchain.persistHeaders(headers);

        for (const header of headers) {
          this._knownHeaderHashes.add(header.hash);
        }
      } finally {
        for (const header of headers) {
          this._tempKnownHeaderHashes.delete(header.hashHex);
        }
      }
    }

    if (this._blockchain.currentHeader.index < peer.data.startHeight) {
      this._sendMessage(peer, this._createMessage({
        command: _Command.COMMAND.GET_HEADERS,
        payload: new _payload.GetBlocksPayload({
          hashStart: [this._blockchain.currentHeader.hash]
        })
      }));
    }
  }

  _onInvMessageReceived(peer, inv) {
    let hashes;

    switch (inv.type) {
      case 0x01:
        // Transaction
        hashes = inv.hashes.filter(hash => !this._knownTransactionHashes.has(hash) && !this._tempKnownTransactionHashes.has(_neoBlockchainCore.common.uInt256ToHex(hash)));
        break;

      case 0x02:
        // Block
        hashes = inv.hashes.filter(hash => !this._knownBlockHashes.has(hash) && !this._tempKnownBlockHashes.has(_neoBlockchainCore.common.uInt256ToHex(hash)));
        break;

      case 0xe0:
        // Consensus
        // TODO: Implement
        hashes = [];
        break;

      default:
        // eslint-disable-next-line
        inv.type;
        hashes = [];
        break;
    }

    if (hashes.length > 0) {
      this._sendMessage(peer, this._createMessage({
        command: _Command.COMMAND.GET_DATA,
        payload: new _payload.InvPayload({
          type: inv.type,
          hashes
        })
      }));
    }
  }

  _onMemPoolMessageReceived(peer) {
    this._sendMessage(peer, this._createMessage({
      command: _Command.COMMAND.INV,
      payload: new _payload.InvPayload({
        type: _payload.INVENTORY_TYPE.TRANSACTION,
        hashes: _neoBlockchainCore.utils.values(this.memPool).map(transaction => transaction.hash)
      })
    }));
  }

  async _onTransactionReceived(transaction) {
    await this.relayTransaction(transaction);
  } // eslint-disable-next-line


  _onVerackMessageReceived(peer) {
    peer.close();
  } // eslint-disable-next-line


  _onVersionMessageReceived(peer) {
    peer.close();
  }

  async _getHeaders(getBlocks, maxHeight) {
    let hashStopIndexPromise = Promise.resolve(maxHeight);

    if (getBlocks.hashStop !== 0) {
      hashStopIndexPromise = this._blockchain.header.tryGet({
        hashOrIndex: getBlocks.hashStop
      }).then(hashStopHeader => hashStopHeader == null ? maxHeight : Math.min(hashStopHeader.index, maxHeight));
    }

    const [hashStartHeaders, hashEnd] = await Promise.all([Promise.all(getBlocks.hashStart.map(hash => this._blockchain.header.tryGet({
      hashOrIndex: hash
    }))), hashStopIndexPromise]);

    const hashStartHeader = _lodash.default.head(_lodash.default.orderBy(hashStartHeaders.filter(Boolean), [header => header.index]));

    if (hashStartHeader == null) {
      return [];
    }

    const hashStart = hashStartHeader.index + 1;

    if (hashStart > maxHeight) {
      return [];
    }

    const headers = await Promise.all(_lodash.default.range(hashStart, Math.min(hashStart + GET_BLOCKS_COUNT, hashEnd)).map(index => this._blockchain.header.get({
      hashOrIndex: index
    })));
    return headers;
  }

  async _trimMemPool() {
    const memPool = _neoBlockchainCore.utils.values(this.memPool);

    if (memPool.length > MEM_POOL_SIZE) {
      const transactionAndFee = await Promise.all(memPool.map(async transaction => {
        const networkFee = await transaction.getNetworkFee({
          getOutput: this._blockchain.output.get,
          governingToken: this._blockchain.settings.governingToken,
          utilityToken: this._blockchain.settings.utilityToken,
          fees: this._blockchain.settings.fees
        });
        return [transaction, networkFee];
      }));

      const hashesToRemove = _lodash.default.take(_lodash.default.sortBy(transactionAndFee, // TODO: Might be a bug since we're converting to number
      ([transaction, networkFee]) => networkFee.divn(transaction.size).toNumber()) // eslint-disable-next-line
      ).map(([transaction, _]) => transaction.hashHex);

      for (const hash of hashesToRemove) {
        delete this.memPool[hash];
      }
    }
  }

  _testFilter(bloomFilterIn, transaction) {
    const bloomFilter = bloomFilterIn;

    if (bloomFilter == null) {
      return true;
    }

    return bloomFilter.contains(transaction.hash) || transaction.outputs.some(output => bloomFilter.contains(output.address)) || transaction.inputs.some(input => bloomFilter.contains(input.serializeWire())) || transaction.scripts.some(script => bloomFilter.contains(_neoBlockchainCore.crypto.toScriptHash(script.verification))) || transaction.type === _neoBlockchainCore.TRANSACTION_TYPE.REGISTER && transaction instanceof _neoBlockchainCore.RegisterTransaction && bloomFilter.contains(transaction.asset.admin);
  }

  _createMerkleBlockPayload({
    block,
    flags
  }) {
    const tree = new _neoBlockchainCore.MerkleTree(block.transactions.map(transaction => transaction.hash)).trim(flags);
    const buffer = Buffer.allocUnsafe(Math.floor((flags.length + 7) / 8));

    for (let i = 0; i < flags.length; i += 1) {
      if (flags[i]) {
        // eslint-disable-next-line
        buffer[Math.floor(i / 8)] |= 1 << i % 8;
      }
    }

    return new _payload.MerkleBlockPayload({
      version: block.version,
      previousHash: block.previousHash,
      merkleRoot: block.merkleRoot,
      timestamp: block.timestamp,
      index: block.index,
      consensusData: block.consensusData,
      nextConsensus: block.nextConsensus,
      script: block.script,
      transactionCount: block.transactions.length,
      hashes: tree.toHashArray(),
      flags: buffer
    });
  }

  _createMessage(value) {
    return new _Message.default({
      magic: this._blockchain.settings.messageMagic,
      value
    });
  }

}

exports.default = Node;