"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.assertInventoryType = exports.InvalidInventoryTypeError = exports.INVENTORY_TYPE = void 0;
const INVENTORY_TYPE = {
  TRANSACTION: 0x01,
  BLOCK: 0x02,
  CONSENSUS: 0xe0
};
exports.INVENTORY_TYPE = INVENTORY_TYPE;

class InvalidInventoryTypeError extends Error {
  constructor(inventoryType) {
    super(`Expected inventory type, found: ${inventoryType}`);
    Object.defineProperty(this, "inventoryType", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    this.inventoryType = inventoryType;
  }

}

exports.InvalidInventoryTypeError = InvalidInventoryTypeError;

const assertInventoryType = inventoryType => {
  switch (inventoryType) {
    case INVENTORY_TYPE.TRANSACTION:
      return INVENTORY_TYPE.TRANSACTION;

    case INVENTORY_TYPE.BLOCK:
      return INVENTORY_TYPE.BLOCK;

    case INVENTORY_TYPE.CONSENSUS:
      return INVENTORY_TYPE.CONSENSUS;

    default:
      throw new InvalidInventoryTypeError(inventoryType);
  }
};

exports.assertInventoryType = assertInventoryType;