"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _neoBlockchainCore = require("neo-blockchain-core");

/* @fl/*  */
class MerkleBlockPayload extends _neoBlockchainCore.BlockBase {
  constructor({
    version,
    previousHash,
    merkleRoot,
    timestamp,
    index,
    consensusData,
    nextConsensus,
    script,
    transactionCount,
    hashes,
    flags
  }) {
    super({
      version,
      previousHash,
      merkleRoot,
      timestamp,
      index,
      consensusData,
      nextConsensus,
      script
    });
    Object.defineProperty(this, "transactionCount", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "hashes", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "flags", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "__merkleBlockPayloadSize", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    this.transactionCount = transactionCount;
    this.hashes = hashes;
    this.flags = flags;
    this.__merkleBlockPayloadSize = _neoBlockchainCore.utils.lazy(() => super.size + _neoBlockchainCore.IOHelper.sizeOfUInt32LE + _neoBlockchainCore.IOHelper.sizeOfArray(this.hashes, () => _neoBlockchainCore.IOHelper.sizeOfUInt256) + _neoBlockchainCore.IOHelper.sizeOfVarBytesLE(this.flags));
  }

  get size() {
    return this.__merkleBlockPayloadSize();
  }

  serializeWireBase(writer) {
    super.serializeWireBase(writer);
    writer.writeVarUIntLE(this.transactionCount);
    writer.writeArray(this.hashes, hash => {
      writer.writeUInt256(hash);
    });
    writer.writeVarBytesLE(this.flags);
  }

  static deserializeWireBase(options) {
    const {
      reader
    } = options;
    const blockBase = super.deserializeBlockBaseWireBase(options);
    const transactionCount = reader.readVarUIntLE(_neoBlockchainCore.utils.INT_MAX_VALUE).toNumber();
    const hashes = reader.readArray(() => reader.readUInt256());
    const flags = reader.readVarBytesLE();
    return new this({
      version: blockBase.version,
      previousHash: blockBase.previousHash,
      merkleRoot: blockBase.merkleRoot,
      timestamp: blockBase.timestamp,
      index: blockBase.index,
      consensusData: blockBase.consensusData,
      nextConsensus: blockBase.nextConsensus,
      script: blockBase.script,
      transactionCount,
      hashes,
      flags
    });
  }

  static deserializeWire(options) {
    return this.deserializeWireBase({
      context: options.context,
      reader: new _neoBlockchainCore.BinaryReader(options.buffer)
    });
  }

}

exports.default = MerkleBlockPayload;