"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _neoBlockchainCore = require("neo-blockchain-core");

class NetworkAddress {
  constructor({
    host,
    port,
    timestamp,
    services
  }) {
    Object.defineProperty(this, "host", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "port", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "timestamp", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "services", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "serializeWire", {
      configurable: true,
      enumerable: true,
      writable: true,
      value: (0, _neoBlockchainCore.createSerializeWire)(this.serializeWireBase.bind(this))
    });
    this.host = host;
    this.port = port;
    this.timestamp = timestamp;
    this.services = services;
  }

  serializeWireBase(writer) {
    // TODO: Support non-ipv4 addresses
    const parts = this.host.split('.');

    if (parts.length !== 4) {
      throw new _neoBlockchainCore.InvalidFormatError();
    }

    const host = Buffer.from(parts.map(part => parseInt(part, 10)));
    writer.writeUInt8(this.timestamp);
    writer.writeUInt64LE(this.services);
    writer.writeBytes(host); // TODO: NEO node reverses endianness of machine, so this is
    //       technically not well-defined. Let's assume it's always
    //       trying to write little endian values.

    writer.writeUInt16LE(this.port);
  }

  static deserializeWireBase({
    reader
  }) {
    const timestamp = reader.readUInt8();
    const services = reader.readUInt64LE();
    const host0 = reader.readUInt8();
    const host1 = reader.readUInt8();
    const host2 = reader.readUInt8();
    const host3 = reader.readUInt8();
    const port = reader.readUInt16LE();
    return new this({
      timestamp,
      services,
      host: [host0, host1, host2, host3].join('.'),
      port
    });
  }

  static deserializeWire(options) {
    return this.deserializeWireBase({
      context: options.context,
      reader: new _neoBlockchainCore.BinaryReader(options.buffer)
    });
  }

}

exports.default = NetworkAddress;